/*
**	branch_anal.c	- Analysis code for branch analyzer
**
**		This file defines the analysis code for the branch analyzer.  The
**		branch analyzer is a simple Atck example that prints statistics
**		about the branch instructions in an application.
*/

#include <atcktarg.h>
#include "branch.h"


/*
 * This prevents the linker from removing all the procedures in this file.
 * Without this, the linker thinks that none of these procedures are called,
 * so it deletes them from the analysis executable.
 */
#pragma force_active on


/*
 * The branch statistics we collect.
 */

brdat_t*		BranchData;



/*
**  Analysis routine called each time a conditional branch is executed.
**  The 'taken' parameter is TRUE if the branch is taken.
*/
void TraceBr(	atck_uint32_t branchID,
				atck_uint32_t predicted,
				atck_uint32_t actual)
{
	BranchData[branchID].nbranches++;
	if (predicted == actual)
		BranchData[branchID].ntaken++;
}

void Initialize(brdat_t* buffer)
{
	BranchData = buffer;
}


/*
**  Analysis routine called when the branch data should be reported.
*/
void Report(void)
{
	/*
	 * We simply stop ourselves and let the host read the branch data.
	 */
	atcktarg_stop();
}
