#include "Candidate.h"

#define CACHE_LINE_LOG2	6
#define CACHE_LINE		(1 << CACHE_LINE_LOG2)

// Arbitrary thresholds for discarding candidates
#define THRESHOLD_CACHE	10

Candidate::Candidate(atck_proc_t* pproc, atck_uint32_t imgID)
{
	atck_bbit_t*  	pbbit;
	atck_instit_t*	piit;
	atck_call_t*  	pcall;
	atck_bb_t*    	pbb_targ;
	atck_bb_t*    	pbb;
	atck_inst_t*  	pi;
	atck_uint32_t	bytes;
	
	// Guilty until proven innocent...
	m_bIsCandidate = false;
	
	// Get file name and line number.
	m_sFile = atck_proc_file(pproc);
	m_uLine = atck_proc_line(pproc);
	
	// If filename or lne number aren't known, then this by definition'
	// isn't a candidate for inlining--how would we go about it without
	// the source file? ;-)
	if (m_sFile == NULL || m_uLine == 0)
		return;

	// Figure out how many cache lines this proc takes up
	bytes = atck_proc_ninst(pproc) * 4;
	m_aAddress = atck_proc_addr(pproc);
	m_uImgID = imgID;
	bytes += m_aAddress & (CACHE_LINE - 1);
	m_uCacheLineCount = (bytes + (CACHE_LINE - 1)) >> CACHE_LINE_LOG2;
	
	if (m_uCacheLineCount > THRESHOLD_CACHE)
		return;
		
	// Check for loops and calls to other procedures.
	m_uCallCount = 0;
	
	pbbit = atck_proc_bbit_new(pproc);
	while (pbb = atck_bbit_next(pbbit))
	{
		piit = atck_bb_instit_new(pbb, ATCK_FLAGS_ITLIFE);
		while (pi = atck_instit_next(piit))
		{
			pcall = atck_inst_callsite(pi);
			if (pcall)
				++m_uCallCount;
			else if (atck_inst_isbranch(pi))
			{
				// If this is a backward branch, we have a loop and
				// are thus not a candidate.

				pbb_targ = atck_inst_branchtarg(pi);
				if (pbb_targ && atck_bb_addr(pbb_targ) <= atck_bb_addr(pbb))
				{
					atck_instit_free(piit);
					atck_bbit_free(pbbit);
					return;
				}
			}
		}
		atck_instit_free(piit);
	}
	atck_bbit_free(pbbit);
	
	// If we made it this far, we passed all the tests!
	m_bIsCandidate = true;
}

void Candidate::Display(void)
{
	printf("Function %s (%s, line %d):", m_sName, m_sFile, m_uLine);
	printf("\tOccupies %d cache lines; makes %d calls.\n", m_uCacheLineCount, m_uCallCount);
}

void CallSite::Display(void)
{
	printf("\tCalled %d times from %s, line %d.\n", m_uCallCount, m_sFile, m_uLine);
}
