#ifndef Candidate_H
#define Candidate_H

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <vector.h>
#include <atck.h>
#include <atcktypes.h>
#include <atckps2.h>

class Candidate
{
public:
	Candidate(atck_proc_t* pproc, atck_uint32_t imgID = 0);
		
	inline ~Candidate()
	{
	}
	
	bool IsCandidate(void) {return m_bIsCandidate;}
	bool DoesAddressMatch(atck_addr_t addr, atck_uint32_t imgID = 0)
	{
		return (m_aAddress == addr && m_uImgID == imgID);
	}
	
	bool operator==(const Candidate& other)
	{
		return DoesAddressMatch(other.m_aAddress, other.m_uImgID);
	}
	
	void Display(void);

protected:
	bool m_bIsCandidate;
	const char* m_sName;
	atck_addr_t m_aAddress;
	atck_uint32_t m_uImgID;
	const char* m_sFile;
	atck_uint32_t m_uLine;
	atck_uint32_t m_uCacheLineCount;
	atck_uint32_t m_uCallCount;
};

class CallSite
{
public:
	CallSite(atck_inst_t* pinst, Candidate* ptarget)
	{
		if (pinst)
		{
			m_sFile = atck_inst_file(pinst);
			m_uLine = atck_inst_line(pinst);
		}
		else
		{
			m_sFile = "[unknown]";
			m_uLine = 0;
		}
		
		m_pTarget = ptarget;
		m_uCallCount = 0;
	}
	
	void Display(void);
	
	void SetCallCount(atck_uint32_t val) {m_uCallCount = val;}
	
	inline ~CallSite()
	{
	}
	
	bool DoesCandidateMatch(Candidate* test)
	{
		return *m_pTarget == *test;
	}

protected:
	const char* m_sFile;
	atck_uint32_t m_uLine;
	Candidate* m_pTarget;
	atck_uint32_t m_uCallCount;
	
friend class CompCallSites;
};

class CompCallSites
{
public:
	bool operator()(const CallSite* p1, const CallSite* p2)
	{
		return (p1->m_pTarget < p2->m_pTarget);
	}
};

typedef vector<CallSite*> CallSiteList;
typedef vector<CallSite*>::iterator CallSiteIterator;
typedef vector<Candidate*>::iterator CandidateIterator;
typedef vector<Candidate*> CandidateList;

#endif	// Candidate_H
