#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <iostream.h>

#include <atck.h>
#include <atcktypes.h>

#include "RunContext.h"

/*
** main
*/
int main (int argc, const char** argv)
{
	RunContext		context;

	/*
	 * Get the command line arguments.
	 */
	if (argc != 2) 
	{
		printf("Usage: inliner <app>\n");
		return 1;
	}

	if (!context.InitSession(argv[1]))
		return 2;
		
	if (!context.BuildCandidateList())
		return 3;
		
	if (!context.Instrument())
		return 4;
		
	if (!context.Run("Press ENTER to finish.\n"))
		return 5;
	
	return 0;
}


















































	
#if 0	
	

/*
** simulate hashtable lookup to find a procedure call
*/
ProcedureCallDetails * pprocLookupProcedureCall (const char * strName)
	{
	// iterate through procedures already stored, searching for strName
	HashProcedureCalls::iterator itProcedures = hashProcedureCalls.begin();
	
	while (itProcedures != hashProcedureCalls.end())
		{
		// test procedure name
		if ((*itProcedures)->bIsNamed(strName))
			{
			return *itProcedures;
			}
			
		// next procedure
		itProcedures++;
		}
		
	return NULL;
	}
	

/*
** test for prohibited procedure calls
*/
void vTestProbihitedProcedures (TrcConfiguration &config)
	{
	TrcProcedureList::iterator itProc = config.m_proclistProhibited.begin();
	
	while (itProc != config.m_proclistProhibited.end())
		{
		TrcProcedure * pprocTestDef = *itProc;
		const char * strProcName = pprocTestDef->strProcedureName();
		ProcedureCallDetails * pprocCall = pprocLookupProcedureCall(strProcName);
	
		printf("TEST - Prohibited function calls to '%s'\n", strProcName);
	
		if (pprocCall != NULL)
			{
			// not allowed! print reason why
			printf(" FAILED!! Calls from:\n", strProcName);
			pprocCall->vPrintCalls();
			printf(" Reason for test: %s\n", pprocTestDef->strReason());
			}
		else
			{
			printf(" PASSED, no calls found.\n");
			}

		printf("\n");
		
		itProc++;
		}
	}


/*
** test for required procedure calls
*/
void vTestRequiredProcedures (TrcConfiguration &config)
	{
	TrcProcedureList::iterator itProc = config.m_proclistRequired.begin();
	
	while (itProc != config.m_proclistRequired.end())
		{
		TrcProcedure * pprocTestDef = *itProc;
		const char * strProcName = pprocTestDef->strProcedureName();
		ProcedureCallDetails * pprocCall = pprocLookupProcedureCall(strProcName);

		printf("TEST - Required function calls to '%s'\n", strProcName);
	
		if (pprocCall == NULL)
			{
			// not allowed! procedure call MUST be present
			// not allowed! print reason why
			printf(" FAILED!!  No function calls found.\n");
			printf(" Reason for test: %s\n", pprocTestDef->strReason());
			}
		else
			{
			// that's good, procedure call present
			printf(" PASSED, call to required procedure from:\n");
			pprocCall->vPrintCalls();
			}

		printf("\n");
				
		itProc++;
		}
	}


/*
** dump call info
*/
void vDumpCallInfo ()
	{
	// iterate through procedures already stored, searching for strName
	HashProcedureCalls::iterator itProcedures = hashProcedureCalls.begin();
	
	while (itProcedures != hashProcedureCalls.end())
		{
		// test procedure name
		(*itProcedures)->vDumpInfo();

		// next procedure
		itProcedures++;
		}	
	}


// utiltity functions to 'mock up' hashtable until we swtich to a real one
void addProcedureCallToHastable (const char * strName, CallSite * pcallsite)
	{
	// iterate through procedures already stored, searching for strName
	HashProcedureCalls::iterator itProcedures = hashProcedureCalls.begin();
	
	while (itProcedures != hashProcedureCalls.end())
		{
		// test procedure name
		if ((*itProcedures)->bIsNamed(strName))
			{
			// found the procedure, add the call
			(*itProcedures)->vAddCall(pcallsite);
			return;		
			}
			
		// next procedure
		itProcedures++;
		}

	// didn't find the procedure so create a new one and add it to the
	// vector
	ProcedureCallDetails * newProc = new ProcedureCallDetails(strName, pcallsite);
	hashProcedureCalls.push_back(newProc);
	}


/*
** dump callsite info
*/
void vDumpCallSite (atck_proc_t * pproc, atck_call_t * pcall, bool bReadOnly)
	{
	// dump target procedure name
	const char * strName = atck_call_targname(pcall);
	if (strName != NULL)
		{
		printf("%s -> %s\n", atck_proc_name(pproc), strName);		
		}
	}


/*
** build data structure of calls to functions
*/
void vAddCallSiteToHashtable (atck_proc_t * pproc, atck_call_t * pcall, bool bReadOnly)
	{
	CallSite * pcallsite = new CallSite(pproc, pcall, bReadOnly);
	
	const char * strName = atck_call_targname(pcall);
	if (strName != NULL)
		{
		addProcedureCallToHastable(strName, pcallsite);
		
		if (g_bVerbose)
			{
			printf(" call %s -> %s\n", atck_proc_name(pproc), strName);
			}
		}
	}


// recursively go through all the calls in the program
void vForeachCallSiteFromMain(atck_prog_t * pprog, 
	void (*fnAction)(atck_proc_t * pproc, atck_call_t * pcall, bool bReadOnly))
	{
	// print out some info on # of call sites
	printf("  %d readable call sites, %d instrumentable\n", 
		atck_ncall_rdable(pprog),
	 	atck_ncall_instr(pprog));
	
	if ((atck_ncall_instr(pprog) == 0) &&
		(g_bUseReadOnly == false))
		{
		printf("WARNING: no instrumentable calls - ELF file is compiled with gcc?\n");
		printf("         Condider running with --use-gcc to include\n");
		printf("         gcc-compiled functions in tests.\n");

		exit(0);
		}

	// lookup start entry point
	// we assume this is 'main'
	atck_ent_t * pent_main = atck_ent_byname(pprog, "main");

	if (pent_main == NULL)
		{
		printf("Can't locate entry point!!\n");
		exit(0);
		}
		
	vRecurseCallSites(pent_main, fnAction); //, &iUnknownCalls);	
	}
	
		
void vRecurseCallSites (atck_ent_t * pent, 
	void (*fnAction)(atck_proc_t * pproc, atck_call_t * pcall, bool bReadOnly))
	{
	// get procedure for entry point
	atck_proc_t * pproc = atck_ent_proc(pent);

	// get attributes for procedure
	atck_attr_t attr = atck_proc_attr(pproc);
	bool readonly = false;

	if (g_bVerbose)
		{
		printf("RECURSE INTO %s:\n", atck_proc_name(pproc));
		}

	// check if we want to count this procedure
	switch (attr) 
		{
		case ATCK_ATTR_INSTR:
			// count it
			break;
			
		case ATCK_ATTR_RDONLY:
			// don't count gcc-complied functions (unless we really want to)
			// don't count SCE library functions
			if (! g_bUseReadOnly)
				{
				if (g_bVerbose) printf(" skipping\n");
				return;
				}
				
			if (g_config.bIsSceFunction(atck_proc_name(pproc)))
				{
				if (g_bVerbose) printf(" skipping SCE function\n");
				return;
				}
				
			if (g_bVerbose) printf(" readonly\n");
			readonly = true;
			break;

		case ATCK_ATTR_NOREAD:
		case ATCK_ATTR_SKIP:
		default:
			// don't count this procedure
			return;
			break;
		}
	
	// iterate thru call sites
	atck_callit_t * pcallit = atck_proc_callit_new(pproc);
	atck_call_t * pcall;
	
	while (pcall = atck_callit_next(pcallit))
		{
		// check if we can determine if the call is taken
		if (atck_call_istargknown(pcall) == ATCK_TRUE)
			{
			// call action function
			fnAction(pproc, pcall, readonly);

			// recurse into function
			vRecurseCallSites(atck_call_targent(pcall), fnAction);
			}
		else
			{
			vRecordUndeterminedCallSite(pproc, pcall);   /**/
			if (g_bVerbose)
				{
				printf(" unknown call %s -> ????\n", atck_proc_name(pproc));
				}
			}
		}

	atck_callit_free(pcallit);
	}
		
// record an undetermined call site
void vRecordUndeterminedCallSite (atck_proc_t *pproc, atck_call_t *pcall)
	{
	// iterate thru call sites
	ListUndeterminedCallSites::iterator itCallSites = listUndeterminedCallSites.begin();
	
	for (; itCallSites != listUndeterminedCallSites.end(); itCallSites++)
		{
		// is this call site already recorded?
		if ((*itCallSites)->bIsCallSite(pcall))
			{
			// yes, do nothing
			return;		
			}
		}

	// didn't find the call site so create a new one and add it to the list
	UndeterminedCallSite * newUnCallSite = new UndeterminedCallSite(pproc, pcall);
	listUndeterminedCallSites.push_back(newUnCallSite);
	}
	
	
// dump out info about undetermined call sites
void vDumpUndeterminedCallSites (void)
	{
	int iSize = listUndeterminedCallSites.size();
	
	if (iSize < 1)
		{
		// no undetermined call sites, good
		return;
		}

	printf("WARNING: %d calls could not have their target determined:\n", iSize);
	
	// iterate thru call sites
	ListUndeterminedCallSites::iterator itCallSites = listUndeterminedCallSites.begin();
	
	for (; itCallSites != listUndeterminedCallSites.end(); itCallSites++)
		{
		UndeterminedCallSite * puncall = *itCallSites;
		
		printf("  Function %s ", puncall->getProcedureName());
		
		if (puncall->bHaveSource())
			{
			printf("\n    %s : %d\n", puncall->getFilename(), puncall->getLineNumber());
			}
		else
			{
			printf("[unknown file, line no]\n");
			}
		}
		
	printf("\n");
	}
	
#endif