#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <atck.h>

static void Analyze(atck_sesn_t*, atck_prog_t*);

int main(
	int    	argc,
	char **	argv)
{
	atck_sesn_t *	psesn;
	atck_cfg_t * 	pcfg;
	atck_prog_t *	pprog;


	if (argc != 2) {
		fprintf(stderr, "Usage: regstall <app>\n");
		return(1);
	}

	/*
	 * Open the input application.
	 */
	psesn = atck_session("1.0", ATCK_FLAGS_NONE);
	pcfg = atck_config_new(psesn, "<syscfg>");
	pprog = atck_open(psesn, argv[1], 0, pcfg, 0);
	if (!pprog)
	{
		atck_endsession(psesn);
		return(1);
	}

	Analyze(psesn, pprog);

	atck_endsession(psesn);
	return(0);
}


static void Analyze(
	atck_sesn_t *	psesn,
	atck_prog_t *	pprog)
{
	atck_imgit_t * 	pimgit;
	atck_procit_t *	pprocit;
	atck_img_t *   	pimg;
	atck_proc_t *  	pproc;
	atck_bbit_t *	pbbit;
	atck_bb_t *  	pbb;
	atck_instit_t*	pinstit;
	atck_instit_t*	poldinstit;
	atck_inst_t*	pinst;
	atck_inst_t*	pprevinst;
	atck_regs_t*	prevwrite;
	atck_regs_t*	curread;
	atck_regs_t*	result;
	atck_uint32_t	line;

	char*         	pbuf_dis;

	pbuf_dis = atck_malloc(psesn, atck_szdisbuf(pprog));
	
	prevwrite = atck_regs_new(pprog);	
	curread = atck_regs_new(pprog);	
	result = atck_regs_new(pprog);	

	pimgit = atck_imgit_new(pprog);
	while (pimg = atck_imgit_next(pimgit))
	{
		pprocit = atck_img_procit_new(pimg);
		while (pproc = atck_procit_next(pprocit))
		{
			pbbit = atck_proc_bbit_new(pproc);
			atck_regs_remall(prevwrite);
			poldinstit = NULL;
			pprevinst = NULL;
			
			while (pbb = atck_bbit_next(pbbit))
			{
				pinstit = atck_bb_instit_new(pbb, ATCK_FLAGS_ITLIFE);
				while (pinst = atck_instit_next(pinstit))
				{
					atck_regs_remall(curread);
					atck_inst_inregs(pinst, curread);
					atck_regs_intset(result, prevwrite, curread);
					if (atck_regs_num(result) > 0)
					{
						line = atck_inst_line(pinst);
						if (line)
						{
							printf("Load delay stall in file %s, line %d:\n", atck_inst_file(pinst), line);
							
							atck_inst_dis(pprevinst, pbuf_dis);
							printf("\t  %s\n", pbuf_dis);

							atck_inst_dis(pinst, pbuf_dis);
							printf("\t> %s\n", pbuf_dis);
						}
					}
					
					atck_regs_remall(prevwrite);
					if (atck_inst_isload(pinst))
						atck_inst_outregs(pinst, prevwrite);
					pprevinst = pinst;
				}
				
				if (poldinstit)
				{
					atck_instit_free(poldinstit);
					poldinstit = NULL;
				}
				poldinstit = pinstit;
			}
			
			if (poldinstit)
			{
				atck_instit_free(poldinstit);
				poldinstit = NULL;
			}
			atck_bbit_free(pbbit);
		}
		atck_procit_free(pprocit);
		
		atck_img_release(pimg);
	}
	atck_imgit_free(pimgit);

	atck_free(psesn, pbuf_dis);
	atck_regs_free(prevwrite);
	atck_regs_free(curread);
	atck_regs_free(result);
}

