/*
**	Plumber		- Simple memory-leak detector
**
**		TODO DESCRIBE
**
**		Usage:
**			plumber <app> TODO ARGS
**
**		Where:
**			<app>          - The application to analyze
**
**		TODO DESCRIBE ARGS
*/

#include <atck.h>
#include <atckps2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shared.h"

#define MEMLOG_MAX	1000

/*
 * Data type to hold information needed to upload data.
 */
typedef struct
{
	atck_sesn_t*	psesn;     		/* atck session */
	atck_prog_t*	pprog;     		/* handle to application */
	atck_addr_t  	addr_bufptr;	/* address of target pointer to profile data */
	atck_addr_t		addr_bufctr;	/* address of target buffer index */
	MemLog*			pbuffer;     	/* host buffer to hold profile data */
	MemLog*			plog;			/* host log to copy buffers into */
	atck_uint32_t	ncount;			/* number of buffer entries fetched */
} RunData;

static atck_iprog_t* Instrument(atck_sesn_t*, atck_prog_t*, MemLog*);
static RunData* Run(atck_sesn_t*, atck_cfg_t*, atck_prog_t*, atck_iprog_t*, MemLog*);
static void HandleEvt(atck_tevt_t*, void*, void*);
static void FetchAndClear(atck_run_t*, RunData*);
static void AnalyzeResults(atck_sesn_t*, atck_prog_t*, MemLog*, atck_uint32_t);

RunData		rundata;

int main(int argc, char** argv)
{
	atck_sesn_t* 	psesn;
	atck_cfg_t*  	pcfg;
	atck_prog_t* 	pprog;
	atck_iprog_t*	piprog;
	char*        	pnam_app;
	MemLog*			pbuffer;
	MemLog*			plog;
	char        	buf[80];
	char        	c;
	RunData*		pcontext;

	/*
	 * Get the command line arguments.
	 */
	if (argc < 2 || argc > 2)
	{
		fprintf(stderr, "Usage: plumber <app>\n");
		return(1);
	}

	pnam_app = argv[1];

	/*
	 * Open the input application and our own tool's analysis file.
	 */
	psesn = atck_session("1.0", ATCK_FLAGS_NONE);
	pcfg = atck_config_new(psesn, "<syscfg>");
	pprog = atck_open(psesn, pnam_app, "<tooldir>..\\Analysis Code\\analcode.elf", pcfg, "<dir><base>_plumber<suf>");
	if (!pprog)
		pprog = atck_open(psesn, pnam_app, "<tooldir>analcode.elf", pcfg, "<dir><base>_plumber<suf>");
	
	if (!pprog)
	{
		atck_endsession(psesn);
		return(1);
	}

	/*
	 * Allocate an array of MemLog structures.  We use this to hold
	 * the profiling data for all procedures.
	 */
	pbuffer = atck_calloc(psesn, MEMLOG_MAX, sizeof(*pbuffer));
	plog = NULL;

	/*
	 * Instrument the application.
	 */
	piprog = Instrument(psesn, pprog, pbuffer);
	if (!piprog)
	{
		atck_free(psesn, pbuffer);
		atck_endsession(psesn);
		return(1);
	}

	printf("Press ENTER to download and run, x-ENTER to quit\n");
	fgets(buf, sizeof(buf), stdin);
	if (buf[0] != 'X' && buf[0] != 'x')
	{
		/*
		 * Run the instrumented application, and print the profile data.
		 */
		pcontext = Run(psesn, pcfg, pprog, piprog, pbuffer);
		if (!pcontext)
		{
			atck_free(psesn, pbuffer);
			atck_endsession(psesn);
			return(1);
		}
		
		printf("ncount = %d\n", pcontext->ncount);
		AnalyzeResults(psesn, pprog, pcontext->plog, pcontext->ncount);
	}

	atck_free(psesn, pbuffer);
	atck_endsession(psesn);
	return(0);
}


/*
**  Instrument the application and return a handle to the instrumented
**  program.  Return NULL on error.
*/
static atck_iprog_t* Instrument(	atck_sesn_t* psesn,
									atck_prog_t* pprog,
									MemLog* pbuffer	)
{
	atck_proto_t* 	pproto_init;
	atck_proto_t* 	pproto_befalloc;
	atck_proto_t* 	pproto_aftalloc;
	atck_proto_t*	pproto_befrel;

	atck_imgit_t* 	pimgit;
	atck_img_t*   	pimg;
	atck_procit_t*	pprocit;
	atck_proc_t*  	pproc;
	atck_callit_t*	pcallit;
	atck_call_t*	pcall;
	const char*			name;
	
	/* callsiteID starts at 1, so that 0 can be used later on to indicate deleted entry */
	atck_uint32_t	callsiteID = 1;
	atck_uint32_t	nummalloc = 0;
	atck_uint32_t	numfree = 0;
	
	/* anything to declare? */
	atck_analtype(psesn, MEMLOG_DECLARE);
	pproto_init = atck_analproto(pprog, "void Initialize(MemLog[],uint32)");
	pproto_befalloc = atck_analproto(pprog, "void BeforeAlloc(regv64,uint32)");
	pproto_aftalloc = atck_analproto(pprog, "void AfterAlloc(regv64)");
	pproto_befrel = atck_analproto(pprog, "void BeforeRelease(regv64,uint32)");

	/* initialize at start of app */
	atck_callbefore(pprog, pproto_init, MEMLOG_MAX, pbuffer, MEMLOG_MAX);
	
	/* iterate through code, adding calls before and after malloc(),
	** and before free().
	*/
	pimgit = atck_imgit_new(pprog);
	while (pimg = atck_imgit_next(pimgit))
	{
		pprocit = atck_img_procit_new(pimg);
		while (pproc = atck_procit_next(pprocit))
		{
			pcallit = atck_proc_callit_new(pproc);
			while (pcall = atck_callit_next(pcallit))
			{
				name = atck_call_targname(pcall);
				if (name != NULL)
				{
					if (!strcmp(name, "malloc"))
					{
						/* GPR4 should be a0 = the num bytes requested */
						/* GPR2 should be v0 = the pointer to the returned memory */
						atck_call_callbefore(pcall, pproto_befalloc, ATCKMIPS_REG_GPR4, callsiteID);
						atck_call_callafter(pcall, pproto_aftalloc, ATCKMIPS_REG_GPR2, callsiteID);
						nummalloc++;
					}
					else if (!strcmp(name, "free"))
					{
						/* GPR4 should be a0 = the memory being freed */
						atck_call_callbefore(pcall, pproto_befrel, ATCKMIPS_REG_GPR4, callsiteID);
						numfree++;
					}
				}
				
				callsiteID++;
			}
			atck_callit_free(pcallit);
		}
		atck_procit_free(pprocit);

		/*
		 * Write out an instrumented version of this image.
		 */
		atck_img_write(pimg);
	}
	atck_imgit_free(pimgit);
	
	printf("Instrumented %d calls to malloc() and %d calls to free().\n", nummalloc, numfree);
	
	/*
	 * Finish writing out the instrumented application.
	 */
	return(atck_finish_write(pprog));
}


/*
**  Run the instrumented program and print the branch statistics when
**  requested.  Return TRUE on success, FALSE on failure.
*/
static RunData* Run(	atck_sesn_t* psesn,
					atck_cfg_t* pcfg,
					atck_prog_t* pprog,
					atck_iprog_t* piprog,
					MemLog* pbuffer	)
{
	atck_dev_t*	pdev;
	atck_run_t*	prun;
	
	char        	buf[80];
	char        	c;


	/* Look up target symbols */
	rundata.addr_bufptr = atck_anal_symaddr(piprog, "TheBuffer");
	rundata.addr_bufctr = atck_anal_symaddr(piprog, "TheBufferIndex");
	
	if (!rundata.addr_bufptr || !rundata.addr_bufctr)
	{
		fprintf(stderr, "Unable to find target symbols.\n");
		return(NULL);
	}
	
	rundata.psesn = psesn;
	rundata.pprog = pprog;
	rundata.pbuffer = pbuffer;
	rundata.plog = NULL;
	rundata.ncount = 0;

	/*
	 * Connect to the target system.  Note, 'pcfg' tells us which system to
	 * connect to.  Events from the target system are handled by HandleEvt()
	 * below.
	 */
	pdev = atck_connect(psesn, pcfg, HandleEvt, &rundata);
	if (!pdev)
		return(NULL);

	/*
	 * Download the instrumented application and start it running.
	 */
	prun = atck_idownload(pdev, piprog, 0, 0);
	atck_continue(prun);

	/*
	 * Stop the application whenever the user presses ENTER.  Either way,
	 * a STOPPED event is sent to the handle_evt() handler below, which causes
	 * the profile information to be printed.
	 */
	printf("Press ENTER to get data, x-ENTER to quit\n");
	for (;;)
	{
		fgets(buf, sizeof(buf), stdin);
		atck_stop(prun);
		if (buf[0] == 'X' || buf[0] == 'x')
			break;
	}

	atck_disconnect(pdev);
	return(&rundata);
}


/*
**  The Atck framework calls this routine whenever there is an event from
**  the target system.
*/
static void HandleEvt(	atck_tevt_t* pevt,
						void* pcontext,
						void* pdummy)
{
	atck_run_t* 	prun;

	switch (atck_tevt_type(pevt))
	{
	case ATCK_TEVT_STOPPED:
		prun = atck_tevt_run(pevt);
		FetchAndClear(prun, (RunData*)pcontext);
		break;

	case ATCK_TEVT_PRINT:
///		printf("%s", atck_tevt_printstr(pevt));
		break;

	case ATCK_TEVT_LOADPCT:
		printf("Downloading... %u%%\n", atck_tevt_loadpct(pevt));
		break;
	}
}

static void FetchAndClear(atck_run_t* prun,	RunData* pcontext)
{
	atck_uint32_t	numentries;
	atck_addr_t		targbuffer;
	atck_uint32_t	bytesread;
	atck_uint32_t	bytestoread;
	MemLog* temp;

	atck_readdata(prun, pcontext->addr_bufctr, &numentries, "uint32", 1);
	atck_readdata(prun, pcontext->addr_bufptr, &targbuffer, "addr", 1);
	atck_readdata(prun, targbuffer, pcontext->pbuffer, "MemLog", numentries);

	bytesread = pcontext->ncount * sizeof(*(pcontext->pbuffer));
	bytestoread = numentries * sizeof(*(pcontext->pbuffer));
	

	if (pcontext->plog == NULL)
		temp = atck_malloc(pcontext->psesn, bytestoread);
	else
		temp = (MemLog*)atck_realloc(pcontext->psesn, pcontext->plog, bytesread + bytestoread);
	pcontext->plog = temp;
	
	memcpy(pcontext->plog + bytesread, pcontext->pbuffer, bytestoread);
	
	pcontext->ncount += numentries;
	
	numentries = 0;
	atck_writedata(prun, pcontext->addr_bufctr, &numentries, "uint32", 1);

	/*
	 * Start the application running again.
	 */
	atck_continue(prun);
}

typedef struct
{
	MemLog*	data;
	atck_uint32_t index;
} SortEntry;

/*
**  Qsort comparison function to sort MemLogs by ptr, then by sequenceID.
*/
static int CompareMemLogsByPtrThenSequenceID(const void* arg1,	const void*	arg2)
{
	atck_addr_t ptr1 = ((const SortEntry*)arg1)->data->ptr;
	atck_addr_t ptr2 = ((const SortEntry*)arg2)->data->ptr;

	atck_uint32_t seq1 = ((const SortEntry*)arg1)->index;
	atck_uint32_t seq2 = ((const SortEntry*)arg2)->index;
	
	if (ptr1 == ptr2)
	{
		if (seq1 > seq2)
			return(1);
		else
			return(-1);
	
	}
	else if (ptr1 > ptr2)
		return(1);
	else
		return(-1);
}

/*
**  Qsort comparison function to sort MemLogs by callsiteID, then by sequenceID.
*/
static int CompareMemLogsByCallsiteIDThenSequenceID(const void* arg1,	const void*	arg2)
{
	atck_uint32_t id1 = ((const SortEntry*)arg1)->data->callsiteID;
	atck_uint32_t id2 = ((const SortEntry*)arg2)->data->callsiteID;

	atck_uint32_t seq1 = ((const SortEntry*)arg1)->index;
	atck_uint32_t seq2 = ((const SortEntry*)arg2)->index;
	
	if (id1 == id2)
	{
		if (seq1 > seq2)
			return(1);
		else
			return(-1);
	
	}
	else if (id1 > id2)
		return(1);
	else
		return(-1);
}

static void AnalyzeResults( atck_sesn_t* psesn,
							atck_prog_t* pprog,
							MemLog* plog,
							atck_uint32_t nentries	)
{
	atck_uint32_t	i;

	atck_uint32_t	blocksallocated = 0;
	atck_uint32_t	blocksfreed = 0;
	atck_uint32_t	bytesallocated = 0;
	atck_uint32_t	bytesfreed = 0;

	atck_imgit_t* 	pimgit;
	atck_img_t*   	pimg;
	atck_procit_t*	pprocit;
	atck_proc_t*  	pproc;
	atck_callit_t*	pcallit;
	atck_call_t*	pcall;
	const char*		name;
	atck_addr_t		tempaddr;
	atck_inst_t*	pinst;
	atck_uint32_t	callsiteID;
	
	SortEntry*		psort;
	psort = atck_malloc(psesn, nentries * sizeof(*psort));
	for (i = 0; i < nentries; i++)
	{
		psort[i].data = &(plog[i]);
		psort[i].index = i;
	}

	/* First, sort allocation list by pointer, then by sequence ID.
	** Result should be pairs of mallocs and frees.
	*/
	
	qsort(psort, nentries, sizeof(*psort), CompareMemLogsByPtrThenSequenceID);

	for (i = 0; i < nentries; i++)
	{
		/*
		** Any malloc not immediately followed by a free is a memory
		** leak.
		*/
		
		if (psort[i].data->ptr != 0)
		{
			bytesallocated += psort[i].data->bytes;
			blocksallocated++;
		}
		
		if ((i + 1) < nentries)
		{
			if (psort[i].data->ptr == psort[i+1].data->ptr)
			{
				if (psort[i+1].data->bytes == 0)
				{
					/* We clear any freed block from the list,
					** incrementing our summary totals.
					*/
					
					bytesfreed += psort[i].data->bytes;
					blocksfreed++;

					psort[i].data->callsiteID = 0;
					psort[i+1].data->callsiteID = 0;
					i++;
				}
			}
		}
		else if (psort[i].data->bytes == 0)
			psort[i].data->callsiteID = 0;
	}
	
	if (bytesallocated <= bytesfreed)
	{
		printf("No leaks detected!\n");
	}
	else
	{
		printf("You have a memory leak!\n");

		/* Now sort the list by callsiteID. */
		
		qsort(psort, nentries, sizeof(*psort), CompareMemLogsByCallsiteIDThenSequenceID);

		/* Now iterate through all callsites. */
		
		callsiteID = 1;
		i = 0;

		pimgit = atck_imgit_new(pprog);
		while (pimg = atck_imgit_next(pimgit))
		{
			pprocit = atck_img_procit_new(pimg);
			while (pproc = atck_procit_next(pprocit))
			{
				pcallit = atck_proc_callit_new(pproc);
				while (pcall = atck_callit_next(pcallit))
				{
					/* Increment i until we reach the current callsiteID. Remember, list is already
					** sorted by callsiteID, so we never need to go backwards!
					** Also remember that matched malloc/free pairs had their callsiteIDs set to 0,
					** which means that we skip them right away.
					*/
					while (i < nentries)
					{
						if (psort[i].data->callsiteID >= callsiteID)
							break;
						i++;
					}
					
					/* If we've finished processing the entire list, then wrap
					** things up--our work is through!
					*/
					if (i >= nentries)
						break;

					name = atck_call_targname(pcall);
					if (name != NULL)
					{
						if (!strcmp(name, "malloc"))
						{
							/* This is a call to malloc(). See if there are any unfreed mallocs() for this
							** site in our list.
							*/
							if (psort[i].data->callsiteID == callsiteID)
							{
								/* Yes, list contains unfreed malloc() here. So get file/line info. */
								printf("Unfreed mallocs at ");
								tempaddr = atck_call_addr(pcall);
								if (tempaddr != NULL)
								{
									pinst = atck_img_inst_byaddr(pimg, tempaddr);
									if (pinst != NULL)
										printf("file %s, line %d", atck_inst_file(pinst), atck_inst_line(pinst));
									else
										printf("unknown location");
								}
								printf(":\n");
								
								do
								{
									if (psort[i].data->callsiteID == callsiteID)
									{
										printf("\t%d bytes at %0x\n", psort[i].data->bytes, psort[i].data->ptr);
									}
									else
										break;
										
									i++;
								} while (i < nentries);
							}
						}
						else if (!strcmp(name, "free"))
						{
							/* This is a call to free(). See if there are any unallocated frees() for this
							** site in our list.
							*/
							
							if (psort[i].data->callsiteID == callsiteID)
							{
								/* Yes, list contains unallocated free() here. So get file/line info. */
								printf("Unallocated frees at ");
								tempaddr = atck_call_addr(pcall);
								if (tempaddr != NULL)
								{
									pinst = atck_img_inst_byaddr(pimg, tempaddr);
									if (pinst != NULL)
										printf("file %s, line %d", atck_inst_file(pinst), atck_inst_line(pinst));
									else
										printf("unknown location");
								}
								printf(":\n");
								
								do
								{
									if (psort[i].data->callsiteID == callsiteID)
									{
										printf("\t%0x\n", psort[i].data->ptr);
									}
									else
										break;
										
									i++;
								} while (i < nentries);
							}
						}
					}
					
					callsiteID++;
				}
				atck_callit_free(pcallit);
				
				/* If we've finished processing the entire list, then wrap
				** things up--our work is through!
				*/
				if (i >= nentries)
					break;
			}
			atck_procit_free(pprocit);
			atck_img_release(pimg);
			
			/* If we've finished processing the entire list, then wrap
			** things up--our work is through!
			*/
			if (i >= nentries)
				break;
		}
		atck_imgit_free(pimgit);
	}
	
	printf("Total allocations: %d bytes in %d blocks.\n", bytesallocated, blocksallocated);
	printf("Total releases: %d bytes in %d blocks.\n", bytesfreed, blocksfreed);
}	
