/* Answer to exercise 1 of Lesson 02. */
atck_inst_t* GetNextInstruction(atck_inst_t* cur)
{
	atck_proc_t* proc;
	atck_bb_t* bb = NULL;
	atck_bb_t* curbb;
	atck_bbit_t* bbit;
	atck_instit_t* instit;
	atck_inst_t* inst;
	atck_inst_t* previnst = NULL;
	atck_inst_t* nextinst;
	
	/* Get the basic block that contains this instruction. */
	curbb = atck_inst_bb(cur);

	/* Create an instruction iterator and position it at the specified
	** instruction. */
	instit = atck_bb_instit_new(curbb, ATCK_FLAGS_NONE);
	do
	{
		previnst = inst;
		inst = atck_instit_next(instit);
	}
	while (inst != NULL && inst != cur)

	if (inst == NULL)
	{
		/* Can't find the original instruction! Something's weird! */
		atck_instit_release(instit);
		return NULL;
	}
	
	/* Get the next instruction within this block. */
	nextinst = atck_instit_next(instit);
	if (nextinst != NULL)
	{
		/* Most of the time that's all it takes. */
		atck_instit_release(instit);
		return nextinst;
	}
	else
	{
		/* This instruction falls at the end of a basic block.
		**	Now we need to determine whether it was this instruction,
		**	or the previous one, that ended the block. */
		
		if (!atck_inst_isbranch(inst))
		{
			/* This instruction is not a branch, so let's see if the
			** previous one was. */
			
			if (!atck_inst_isbranch(previnst))
			{
				/* If neither instruction is a branch, then the block ended
				** because some other code branched into the middle of the
				** block.
				** Therefore, just return first instruction of next block. */
				
				atck_instit_release(instit);

				/* Step through the parent procedure's blocks to find this one. */
				proc = atck_bb_proc(curbb);
				bbit = atck_proc_bbit_new(proc);
				do
					bb = atck_bbit_next(bbit);
				while (bb != NULL && bb != curbb)

				if (bb == NULL)
				{
					/* Can't find the current block! Something's weird! */
					atck_bbit_release(bbit);
					return NULL;
				}
				else
				{
					/* The iterator is pointing to the current block.
					** Increment it to point to the next block. */
					
					bb = atck_bbit_next(bbit);
					atck_bbit_release(bbit);
					if (bb == NULL)
					{
						/* This instruction fell at the end of a procedure with
						** no return. Weird! */
						return NULL;
					}
					else
					{
						/* Now just return the first instruction of this block. */
						instit = atck_bb_instit_new(bb, ATCK_FLAGS_NONE);
						nextinst = atck_instit_next(instit);
						atck_instit_release(instit);
						return nextinst;
					}
				}
			}
			else
			{
				/* inst wasn't a branch, but previnst was. For convenience,
				** copy previnst into inst and continue. */
				inst = previnst;
			}
		}
		
		/* If we've gotten here, we know that inst is a branch to the
		** next instruction. */
		bb = atck_inst_branchtarg(inst);
		if (bb == NULL)
		{
			/* Destination not known! */
			return NULL;
		}
		else
		{
			/* Now just return the first instruction of the target block. */
			instit = atck_bb_instit_new(bb, ATCK_FLAGS_NONE);
			nextinst = atck_instit_next(instit);
			atck_instit_release(instit);
			return nextinst;
		}
	}
}
