#include <stdio.h>
#include <atck.h>

#define MAXPROCS	1000

main(int argc, char **argv)
{
    atck_sesn_t * 	psesn;
    atck_cfg_t *		pcfg;
    atck_prog_t * 	pprog;
    atck_proto_t *	pproto_pcnt;
    atck_imgit_t *	pimgi;
    atck_img_t *  	pimg;
    atck_procit_t *	ppi;
    atck_proc_t * 	pproc;
    atck_iprog_t *	piprog;
    atck_dev_t *   	pdev;
    atck_run_t *  	prun;
    atck_addr_t 		addr;
    atck_uint32_t 	proc_counts[MAXPROCS];
    atck_uint32_t 	iproc;


    /*
     * Open the application.
     */
    psesn = atck_session(ATCK_FLAGS_NONE);
    pcfg = atck_config_new(psesn, "<syscfg>");
    pprog = atck_open(psesn, argv[1], "<tooldir>pcount.anal", pcfg, 0);

    /*
     * Make sure we don't exceed the buffer in the analysis code.
     */
    if (atck_nproc_instr(pprog) > MAXPROCS) {
        fprintf(stderr, "Too many procedures in application.\n");
        atck_endsession(psesn);
        return(1);
    }

    /*
     * Declare the Pcount() analysis routine.
     */
    pproto_pcnt = atck_analproto(pprog, "void Pcount(uint32)");

    /*
     * Make an instrumentation call to Pcount() at the start of every
     * procedure.
     */
    iproc = 0;
    pimgi = atck_imgit_new(pprog);
    pimg = atck_imgit_first(pimgi);
    while (pimg) {
        ppi = atck_img_procit_new(pimg);
        pproc = atck_procit_first(ppi);
        while (pproc) {
            if (atck_proc_attr(pproc) == ATCK_ATTR_INSTR)
                atck_proc_callbefore(pproc, pproto_pcnt, iproc++);
            pproc = atck_procit_next(ppi);
        }
        atck_procit_free(ppi);

        atck_img_write(pimg);
        pimg = atck_imgit_next(pimgi);
    }
    atck_imgit_free(pimgi);

    piprog = atck_finish_write(pprog);

    /*
     * Connect to the target system, and run the instrumented
     * application.
     */
    pdev = atck_connect(psesn, pcfg, 0, 0);
    prun = atck_idownload(pdev, piprog, 0, 0);
    atck_continue(prun);

    /*
     * Stop the application when the user is ready.
     */
    printf("Press ENTER to get data\n");
    getchar();
    atck_stop(prun);

    /*
     * Upload the data.
     */
    addr = atck_anal_symaddr(piprog, "ProcCounts");
    atck_readdata(prun, addr, proc_counts, "uint32", MAXPROCS);

    iproc = 0;
    pimgi = atck_imgit_new(pprog);
    pimg = atck_imgit_first(pimgi);
    while (pimg) {
        ppi = atck_img_procit_new(pimg);
        pproc = atck_procit_first(ppi);
        while (pproc) {
            if (atck_proc_attr(pproc) == ATCK_ATTR_INSTR) {
                if (proc_counts[iproc]) {
                    printf("Procedure '%s' called %u times.\n",
                        atck_proc_name(pproc),
                        (unsigned)proc_counts[iproc]);
                }
                iproc++;
            }
            pproc = atck_procit_next(ppi);
        }
        atck_procit_free(ppi);
        pimg = atck_imgit_next(pimgi);
    }
    atck_imgit_free(pimgi);

    atck_endsession(psesn);
    return(0);
}
